/*
 * Copyright 2016 (c) Anna Schumaker.
 */
#include <core/audio.h>
#include <core/playlist.h>
#include <core/settings.h>
#include <gui/builder.h>
#include <gui/filter.h>
#include <gui/queue.h>
#include <gui/view.h>
#include <gui/window.h>
#include <glib/gi18n.h>
#include <stdlib.h>

static const gchar *QUEUE_SETTINGS[GUI_MODEL_N_COLUMNS] = {
	[GUI_MODEL_TRACK_NR]  = "gui.queue.track",
	[GUI_MODEL_TITLE]     = "gui.queue.title",
	[GUI_MODEL_LENGTH]    = "gui.queue.length",
	[GUI_MODEL_ARTIST]    = "gui.queue.artist",
	[GUI_MODEL_ALBUM]     = "gui.queue.album",
	[GUI_MODEL_YEAR]      = "gui.queue.year",
	[GUI_MODEL_GENRE]     = "gui.queue.genre",
	[GUI_MODEL_COUNT]     = "gui.queue.count",
};

static const enum compare_t QUEUE_SORT[GUI_MODEL_N_COLUMNS] = {
	[GUI_MODEL_TRACK_NR]  = COMPARE_TRACK,
	[GUI_MODEL_TITLE]     = COMPARE_TITLE,
	[GUI_MODEL_LENGTH]    = COMPARE_LENGTH,
	[GUI_MODEL_ARTIST]    = COMPARE_ARTIST,
	[GUI_MODEL_ALBUM]     = COMPARE_ALBUM,
	[GUI_MODEL_YEAR]      = COMPARE_YEAR,
	[GUI_MODEL_GENRE]     = COMPARE_GENRE,
	[GUI_MODEL_COUNT]     = COMPARE_COUNT,
	[GUI_MODEL_LAST_PLAY] = COMPARE_PLAYED,
};

static GtkTreeView *view_treeview = NULL;
static unsigned int view_sort_count = 0;
static bool view_no_scroll = false;

static unsigned int __view_get_column_index(GtkTreeViewColumn *col)
{
	unsigned int i;

	if (!view_treeview)
		return GUI_MODEL_N_COLUMNS;

	for (i = 0; i < GUI_MODEL_N_COLUMNS; i++) {
		if (col == gtk_tree_view_get_column(view_treeview, i))
			return i;
	}

	return GUI_MODEL_N_COLUMNS;
}

static inline void __view_display_sorting(gchar *text)
{
	gtk_label_set_text(GTK_LABEL(gui_builder_widget("o_sorting")), text);
}

static int __view_dec_sort(gpointer data)
{
	if (view_sort_count > 0)
		view_sort_count--;
	if (view_sort_count == 0)
		__view_display_sorting("");
	return FALSE;
}

static void __view_set_column_sort_indicator(GtkTreeViewColumn *col,
					     unsigned int index)
{
	struct playlist *playlist = gui_model_get_playlist();
	GSList *cur = playlist ? playlist->pl_queue.q_sort : NULL;
	unsigned int order = GTK_SORT_ASCENDING;
	bool show = false;
	int field;

	while (cur) {
		order = GTK_SORT_ASCENDING;
		field = GPOINTER_TO_INT(cur->data);

		if (abs(field) == QUEUE_SORT[index]) {
			show = true;
			if (field < 0)
				order = GTK_SORT_DESCENDING;
			break;
		}

		cur = g_slist_next(cur);
	}

	gtk_tree_view_column_set_sort_indicator(col, show);
	gtk_tree_view_column_set_sort_order(col, order);
}

static void __view_set_sort_indicators()
{
	GtkTreeViewColumn *col;

	for (unsigned int i = 0; i < GUI_MODEL_N_COLUMNS; i++) {
		col = gtk_tree_view_get_column(view_treeview, i);
		if (col)
			__view_set_column_sort_indicator(col, i);
	}
}

void __view_row_activated(GtkTreeView *treeview, GtkTreePath *path,
			  GtkTreeViewColumn *col, gpointer data)
{
	view_no_scroll = true;
	audio_load(gui_filter_path_get_track(path));
	queue_selected(&gui_model_get_playlist()->pl_queue,
		       gtk_tree_path_get_indices(path)[0]);
	view_no_scroll = false;
}

void __view_column_resized(GtkTreeViewColumn *col, GParamSpec *pspec,
			   gpointer data)
{
	unsigned int index = __view_get_column_index(col);

	settings_set(QUEUE_SETTINGS[index], gtk_tree_view_column_get_width(col));
}

void __view_column_clicked(GtkTreeViewColumn *col, gpointer data)
{
	struct playlist *playlist = gui_model_get_playlist();
	struct queue *queue = playlist ? &playlist->pl_queue : NULL;
	unsigned int index  = __view_get_column_index(col);
	bool reset = view_sort_count == 0;
	gchar *text;

	if (!queue || queue_has_flag(queue, Q_NO_SORT))
		return;

	if (playlist_get_queue(PL_SYSTEM, gui_queue(queue)->gq_text))
		playlist_sort(PL_SYSTEM, gui_queue(queue)->gq_text, QUEUE_SORT[index], reset);
	else
		queue_sort(queue, QUEUE_SORT[index], reset);

	if (view_sort_count == 0) {
		text = g_strdup_printf("Sorting within %s",
					gtk_tree_view_column_get_title(col));
		__view_display_sorting(text);
		g_free(text);
	}

	__view_set_sort_indicators();

	view_sort_count++;
	g_timeout_add_seconds(3, __view_dec_sort, NULL);
}

struct view_add_data {
	enum playlist_type_t vad_type;
	const gchar         *vad_name;
};

static void __view_add_to_playlist(GtkTreeModel *model, GtkTreePath *path,
				   GtkTreeIter  *iter,  gpointer data)
{
	struct view_add_data *vad = (struct view_add_data *)data;
	playlist_add(vad->vad_type, vad->vad_name, gui_filter_path_get_track(path));
}

static void __view_delete_selection(GtkTreeSelection *selection)
{
	struct playlist *playlist = gui_model_get_playlist();
	struct queue *queue = playlist ? &playlist->pl_queue : NULL;
	GList *rows = gtk_tree_selection_get_selected_rows(selection, NULL);
	GList *cur  = g_list_reverse(rows);

	while (cur) {
		queue_erase(queue, gui_filter_path_get_index(cur->data));
		cur = g_list_next(cur);
	}

	g_list_free_full(rows, (GDestroyNotify) gtk_tree_path_free);
}

static gchar *__view_get_new_playlist_name(void)
{
	unsigned int flags = GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL;
	GtkWidget *entry, *dialog, *content;
	gchar *text = NULL;

	entry  = gtk_entry_new();
	dialog = gtk_dialog_new_with_buttons("New Playlist Name?",
					     gui_window(), flags,
					     _("_Cancel"), GTK_RESPONSE_CANCEL,
					     _("_OK"),     GTK_RESPONSE_ACCEPT,
					     NULL);
	content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));

	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), true);
	gtk_container_add(GTK_CONTAINER(content), entry);
	gtk_widget_show_all(dialog);
	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
		text = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	gtk_widget_destroy(dialog);

	return text;
}

static void __view_process_selection(GtkTreeView *treeview, unsigned int keyval)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
	struct view_add_data vad_data;
	gchar *text = NULL;

	switch (keyval) {
	case GDK_KEY_f:
		vad_data.vad_type = PL_SYSTEM;
		vad_data.vad_name = "Favorites";
		gtk_tree_selection_selected_foreach(selection,
						__view_add_to_playlist,
						&vad_data);
		break;
	case GDK_KEY_p:
		text = __view_get_new_playlist_name();
		if (!text || !playlist_new(PL_USER, text))
			break;
		vad_data.vad_type = PL_USER;
		vad_data.vad_name = text;
		gtk_tree_selection_selected_foreach(selection,
						__view_add_to_playlist,
						&vad_data);
		g_free(text);
		break;
	case GDK_KEY_q:
		vad_data.vad_type = PL_SYSTEM;
		vad_data.vad_name = "Queued Tracks";
		gtk_tree_selection_selected_foreach(selection,
						__view_add_to_playlist,
						&vad_data);
		break;
	case GDK_KEY_Delete:
		__view_delete_selection(selection);
		break;
	}
}

void __view_keypress(GtkTreeView *treeview, GdkEventKey *event, gpointer data)
{
	__view_process_selection(treeview, event->keyval);
}

void __view_rc_new_queue(GtkMenuItem *item, gpointer data)
{
	__view_process_selection(view_treeview, GDK_KEY_q);
}

void __view_rc_add_to_queue(GtkMenuItem *item, gpointer data)
{
	unsigned int i;
	gchar *name;

	for (i = 0; i < 10; i++) {
		name = g_strdup_printf("o_queue_%d", i);
		if (GTK_WIDGET(item) == gui_builder_widget(name))
			__view_process_selection(view_treeview, GDK_KEY_0 + i);
		g_free(name);
	}
}

void __view_rc_add_favorites(GtkMenuItem *item, gpointer data)
{
	__view_process_selection(view_treeview, GDK_KEY_f);
}

void __view_rc_add_hidden(GtkMenuItem *item, gpointer data)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection(view_treeview);
	GList *rows = gtk_tree_selection_get_selected_rows(selection, NULL);
	GList *cur  = g_list_reverse(rows);

	while (cur) {
		playlist_add(PL_SYSTEM, "Hidden", gui_filter_path_get_track(cur->data));
		cur = g_list_next(cur);
	}

	g_list_free_full(rows, (GDestroyNotify) gtk_tree_path_free);

}

void __view_rc_add_new(GtkMenuItem *item, gpointer data)
{
	__view_process_selection(view_treeview, GDK_KEY_p);
}

static void __view_rc_add_other(GtkWidget *item, gpointer data)
{
	GtkTreeView      *treeview  = GTK_TREE_VIEW(gui_builder_widget("o_treeview"));
	GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
	struct playlist  *playlist  = (struct playlist *)data;
	struct view_add_data vad_data = {
		.vad_type = PL_USER,
		.vad_name = playlist->pl_name,
	};

	gtk_tree_selection_selected_foreach(selection,
					    __view_add_to_playlist,
					    &vad_data);
}

static gint __view_cmp_pl(gconstpointer a, gconstpointer b)
{
	struct playlist *pl_a = (struct playlist *)a;
	struct playlist *pl_b = (struct playlist *)b;
	return g_utf8_collate(pl_a->pl_name, pl_b->pl_name);
}

static GtkWidget *__view_rc_build_submenu(void)
{
	struct db_entry *dbe, *next;
	struct playlist *playlist;
	GtkWidget *submenu, *item;
	GList *iter, *list = NULL;

	if (pl_user_db_get()->db_size == 0)
		return NULL;

	db_for_each(dbe, next, pl_user_db_get()) {
		playlist = &USER_PLAYLIST(dbe)->pl_playlist;
		list     = g_list_insert_sorted(list, playlist, __view_cmp_pl);
	}

	submenu = gtk_menu_new();
	iter    = g_list_first(list);
	while (iter) {
		playlist = (struct playlist *)iter->data;
		item     = gtk_menu_item_new_with_label(playlist->pl_name);
		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), item);
		g_signal_connect(item, "activate",
				 G_CALLBACK(__view_rc_add_other), playlist);
		iter = g_list_next(iter);
	}

	return submenu;
}

bool __view_button_press(GtkTreeView *treeview, GdkEventButton *event,
			 gpointer data)
{
	GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
	GtkMenu *menu = GTK_MENU(gui_builder_widget("o_menu"));
	GtkWidget *submenu = NULL;
	GtkMenuItem *other;
	GtkTreePath *path;

	if (event->button != GDK_BUTTON_SECONDARY)
		return false;

	/* Select path if it isn't already selected */
	if (gtk_tree_view_get_path_at_pos(treeview, event->x, event->y,
					  &path, NULL, NULL, NULL))
	{
		gtk_tree_selection_select_path(selection, path);
		gtk_tree_path_free(path);
	}

	/* Set the "Other Playlists" submenu. */
	other   = GTK_MENU_ITEM(gui_builder_widget("o_add_to_other"));
	submenu = __view_rc_build_submenu();
	gtk_menu_item_set_submenu(other, submenu);

	if (!submenu)
		gtk_widget_hide(GTK_WIDGET(other));
	else
		gtk_widget_show_all(GTK_WIDGET(other));

	gtk_menu_popup_at_pointer(menu, (GdkEvent *)event);
	return true;
}

void gui_view_init()
{
	GtkTreeViewColumn *col;
	int i, pos;

	view_treeview = GTK_TREE_VIEW(gui_builder_widget("o_treeview"));

	for (i = 0; i < GUI_MODEL_N_COLUMNS; i++) {
		col = gtk_tree_view_get_column(view_treeview, i);
		pos = settings_get(QUEUE_SETTINGS[i]);
		if (col && pos > 0)
			gtk_tree_view_column_set_fixed_width(col, pos);
	}

	gtk_tree_view_set_model(view_treeview, GTK_TREE_MODEL(gui_filter_get()));
}

void gui_view_set_playlist(struct playlist *playlist)
{
	gui_filter_set_playlist(playlist);

	view_sort_count = 0;
	__view_display_sorting("");
	__view_set_sort_indicators();

	gui_view_scroll();
}

void gui_view_scroll()
{
	struct playlist *playlist = gui_model_get_playlist();
	struct queue *queue = playlist ? &playlist->pl_queue : NULL;
	GtkTreePath *path;

	if (!queue || (int)queue->q_cur.it_pos < 0 || view_no_scroll)
		return;

	path = gui_filter_path_from_index(queue->q_cur.it_pos);
	if (!path)
		return;

	gtk_tree_view_set_cursor(view_treeview, path, NULL, false);
	gtk_tree_view_scroll_to_cell(view_treeview, path, NULL, TRUE, 0.5, 0.5);
	gtk_tree_path_free(path);
}
